import pandas as pd

from scipy import stats

import dtale.global_state as global_state
import dtale.pandas_util as pandas_util
from dtale.query import run_query
from dtale.utils import make_list


def flatten_columns(df, columns=None):
    if columns is not None:
        return [
            " ".join(
                [
                    "{}-{}".format(c1, str(c2))
                    for c1, c2 in zip(make_list(columns), make_list(col_val))
                ]
            ).strip()
            for col_val in df.columns.values
        ]
    return [
        " ".join([str(c) for c in make_list(col)]).strip() for col in df.columns.values
    ]


class DataReshaper(object):
    def __init__(self, data_id, shape_type, cfg):
        self.data_id = data_id
        if shape_type == "pivot":
            self.builder = PivotBuilder(cfg)
        elif shape_type == "aggregate":
            self.builder = AggregateBuilder(cfg)
        elif shape_type == "transpose":
            self.builder = TransposeBuilder(cfg)
        elif shape_type == "resample":
            self.builder = ResampleBuilder(cfg)
        else:
            raise NotImplementedError(
                "{} data re-shaper not implemented yet!".format(shape_type)
            )

    def reshape(self):
        data = run_query(
            global_state.get_data(self.data_id),
            global_state.get_query(self.data_id),
            global_state.get_context_variables(self.data_id),
        )
        return self.builder.reshape(data)

    def build_code(self):
        return self.builder.build_code()


class PivotBuilder(object):
    def __init__(self, cfg):
        self.cfg = cfg

    def reshape(self, data):
        index, columns, values, aggfunc = (
            self.cfg.get(p) for p in ["index", "columns", "values", "aggfunc"]
        )
        pivot_data = pd.pivot_table(
            data, values=values, index=index, columns=columns, aggfunc=aggfunc
        )
        if len(values) == 1:
            pivot_data.columns = pivot_data.columns.droplevel(0)
        if self.cfg.get("columnNameHeaders", False):
            pivot_data.columns = flatten_columns(pivot_data, columns=columns)
        else:
            pivot_data.columns = flatten_columns(pivot_data)
        pivot_data = pivot_data.rename_axis(None, axis=1)
        return pivot_data

    def build_code(self):
        index, columns, values, aggfunc = (
            self.cfg.get(p) for p in ["index", "columns", "values", "aggfunc"]
        )
        code = []
        if aggfunc is not None or len(values) > 1:
            code.append(
                "df = pd.pivot_table(df, index='{}', columns='{}', values=['{}'], aggfunc='{}')".format(
                    index, columns, "', '".join(values), aggfunc
                )
            )
            if len(values) > 1:
                code.append(
                    "df.columns = [' '.join([str(c) for c in col]).strip() for col in df.columns.values]"
                )
            elif len(values) == 1:
                code.append("df.columns = df.columns.droplevel(0)")
        else:
            code.append(
                "df = df.pivot(index='{index}', columns='{columns}', values='{values}')".format(
                    index=index, columns=columns, values=values[0]
                )
            )
        code.append("df = df.rename_axis(None, axis=1)")
        return "\n".join(code)


def str_joiner(vals, join_char="|"):
    return join_char.join(vals)


def custom_agg_handler(agg):
    if agg == "gmean":
        return stats.gmean
    if agg == "str_joiner":
        return str_joiner
    return agg


def custom_aggregate_handler(cols):
    return {
        col: [custom_agg_handler(agg) for agg in aggs] for col, aggs in cols.items()
    }


def custom_str_handler(aggs):
    def _handler():
        for agg in aggs:
            if agg == "gmean":
                yield agg
            elif agg == "str_joiner":
                yield "'|'.join"
            else:
                yield "'{}'".format(agg)

    return list(_handler())


class AggregateBuilder(object):
    def __init__(self, cfg):
        self.cfg = cfg

    def reshape(self, data):
        index, agg, dropna = (self.cfg.get(p) for p in ["index", "agg", "dropna"])
        agg_type, func, cols = (agg.get(p) for p in ["type", "func", "cols"])

        if index:
            agg_data = pandas_util.groupby(
                data, index, dropna=dropna if dropna is not None else True
            )
            if agg_type == "func":
                if "count_pct" == func:
                    counts = agg_data.size()
                    return pd.DataFrame(
                        {"Count": counts, "Percentage": (counts / len(data)) * 100}
                    )
                if cols:
                    agg_data = agg_data[cols]
                elif pandas_util.is_pandas2():
                    non_str_cols = [
                        c
                        for c in data.select_dtypes(exclude="object").columns
                        if hasattr(agg_data, c) and c != index
                    ]
                    agg_data = agg_data[non_str_cols]

                return (
                    agg_data.agg(stats.gmean)
                    if func == "gmean"
                    else getattr(agg_data, func)()
                )
            agg_data = agg_data.aggregate(custom_aggregate_handler(cols))
            agg_data.columns = flatten_columns(agg_data)
            return agg_data

        agg_data = data[cols] if cols else data
        if agg_type == "func":
            agg_data = (
                agg_data.apply(stats.gmean)
                if func == "gmean"
                else getattr(agg_data, func)()
            )
            return agg_data.to_frame().T

        agg_data = agg_data.aggregate(custom_aggregate_handler(cols))
        agg_data = agg_data.to_frame().T
        return agg_data

    def build_code(self):
        index, agg, dropna = (self.cfg.get(p) for p in ["index", "agg", "dropna"])
        dropna = dropna if dropna is not None else True
        agg_type, func, cols = (agg.get(p) for p in ["type", "func", "cols"])
        code = []
        if (agg_type == "func" and func == "gmean") or (
            agg_type != "func" and "gmean" in cols.values()
        ):
            code.append("\nfrom scipy.stats import gmean\n\n")

        if index:
            index = "', '".join(index)
            if agg_type == "func":
                if "count_pct" == agg:
                    code.append(
                        (
                            "total_records = len(df)\n"
                            "df = df{groupby}\n"
                            "counts = df.size()\n"
                            "df = pd.DataFrame({'Count': counts, 'Percentage': (counts / total_records) * 100})"
                        )
                    )
                    return code
                agg_str = ".agg(gmean)" if agg == "gmean" else ".{}()".format(agg)
                if cols is not None:
                    code.append(
                        "df = df{groupby}['{columns}']{agg}".format(
                            groupby=pandas_util.groupby_code(index, dropna=dropna),
                            columns="', '".join(cols),
                            agg=agg_str,
                        )
                    )
                    return code
                code.append(
                    "df = df{groupby}{agg}".format(
                        groupby=pandas_util.groupby_code(index, dropna=dropna),
                        agg=agg_str,
                    )
                )
                return code
            code += [
                "df = df{groupby}.aggregate(".format(
                    groupby=pandas_util.groupby_code(index, dropna=dropna)
                )
                + "{",
                ",\n".join(
                    "\t'{col}': ['{aggs}']".format(
                        col=col, aggs=", ".join(custom_str_handler(aggs))
                    )
                    for col, aggs in cols.items()
                ),
                "})",
                "df.columns = [' '.join([str(c) for c in col]).strip() for col in df.columns.values]",
            ]
            return "\n".join(code)

        if cols:
            code.append("df = df[[{}]]".format("', '".join(cols)))
        if agg_type == "func":
            agg_str = ".apply(gmean)" if agg == "gmean" else ".{}()".format(agg)
            code += ["df = df{}".format(agg_str), "df = df.to_frame().T"]
            return code
        code += [
            "df = df.aggregate({"
            + ",\n".join(
                "\t'{col}': ['{aggs}']".format(
                    col=col, aggs=", ".join(custom_agg_handler(aggs))
                )
                for col, aggs in cols.items()
            )
            + "})",
            "df = df.to_frame().T",
        ]
        return code


class TransposeBuilder(object):
    def __init__(self, cfg):
        self.cfg = cfg

    def reshape(self, data):
        index, columns = (self.cfg.get(p) for p in ["index", "columns"])
        t_data = data.set_index(index)
        if any(t_data.index.duplicated()):
            raise Exception(
                "Transposed data contains duplicates, please specify additional index or filtering"
            )
        if columns is not None:
            t_data = t_data[columns]
        t_data = t_data.T
        if len(index) > 1:
            t_data.columns = flatten_columns(t_data)
        t_data = t_data.rename_axis(None, axis=1)
        return t_data

    def build_code(self):
        index, columns = (self.cfg.get(p) for p in ["index", "columns"])

        code = []
        if columns is not None:
            code.append(
                "df = df.set_index('{}')['{}'].T".format(
                    "', '".join(index), "', '".join(columns)
                )
            )
        else:
            code.append("df = df.set_index('{}').T".format("', '".join(index)))
        if len(index) > 1:
            code.append(
                "df.columns = [' '.join([str(c) for c in col]).strip() for col in df.columns.values]"
            )
        code.append("df = df.rename_axis(None, axis=1)")
        return "\n".join(code)


class ResampleBuilder(object):
    def __init__(self, cfg):
        self.cfg = cfg

    def reshape(self, data):
        index, columns, freq, agg = (
            self.cfg.get(p) for p in ["index", "columns", "freq", "agg"]
        )
        t_data = data.set_index(index)
        if columns is not None:
            t_data = t_data[columns]
        t_data = getattr(t_data.resample(freq), agg)()
        if not columns or len(columns) > 1:
            t_data.columns = flatten_columns(t_data)
        t_data.index.name = "{}_{}".format(index, freq)
        t_data = t_data.reset_index()
        return t_data

    def build_code(self):
        index, columns, freq, agg = (
            self.cfg.get(p) for p in ["index", "columns", "freq", "agg"]
        )
        code = []
        if columns is not None:
            code.append(
                "df = df.set_index('{}')['{}'].resample('{}').{}()".format(
                    index, "', '".join(columns), freq, agg
                )
            )
        else:
            code.append(
                "df = df.set_index('{}').resample('{}').{}()".format(index, freq, agg)
            )
        if not columns or len(columns) > 1:
            code.append(
                "df.columns = [' '.join([str(c) for c in col]).strip() for col in df.columns.values]"
            )
        return "\n".join(code)



import dash_bootstrap_components as dbc

from dash.dependencies import Input, Output, State

from dtale.charts.utils import ANIMATION_CHARTS, ANIMATE_BY_CHARTS, ZAXIS_CHARTS
from dtale.dash_application import dcc, html
from dtale.dash_application.charts import build_chart, valid_chart
from dtale.dash_application.exceptions import DtalePreventUpdate
from dtale.dash_application.layout.utils import build_hoverable
from dtale.utils import dict_merge, is_app_root_defined, flatten_lists, make_list
from dtale.translations import text

MAX_SAVED_CHARTS = 10
SAVED_CHART_IDS = list(range(1, MAX_SAVED_CHARTS + 1))


def build_layout():
    return flatten_lists(
        [
            [
                dcc.Store(id="saved-chart-config-{}".format(i)),
                dcc.Store(id="prev-saved-chart-config-{}".format(i)),
                dcc.Store(id="saved-deletes-{}".format(i), data=0),
                html.Div(
                    [
                        html.Div(
                            [
                                html.H3(
                                    "{} {}".format(text("Saved Chart"), i),
                                    className="col-auto pr-3",
                                ),
                                html.Div(
                                    id="saved-chart-header-{}".format(i),
                                    className="col pl-0",
                                ),
                                html.Div(
                                    dbc.Button(
                                        text("Delete"),
                                        id="delete-saved-btn-{}".format(i),
                                        color="primary",
                                        className="delete-chart",
                                    ),
                                    className="col-auto",
                                ),
                            ],
                            className="row",
                        ),
                        html.Div(id="saved-chart-{}".format(i)),
                    ],
                    id="saved-chart-div-{}".format(i),
                    className="saved-chart-div pt-5",
                    style=dict(display="none"),
                ),
            ]
            for i in SAVED_CHART_IDS
        ]
    )


def build_saved_header(config):
    chart_type = config["chart_type"]
    final_data = [
        ("Data ID", config["data_id"]),
        ("Query", config.get("query")),
        ("Chart Type", chart_type),
    ]
    if config.get("agg") not in [None, "raw"]:
        final_data.append(("Aggregation", config["agg"]))

    if chart_type == "maps":
        group_by = config["map_group"]
        map_type = config.get("map_type")
        if map_type == "scattergeo":
            map_props = ["map_type", "lat", "lon", "map_val", "scope", "proj"]
        elif map_type == "mapbox":
            map_props = ["map_type", "lat", "lon", "map_val", "mapbox_style"]
        else:
            map_props = ["map_type", "loc_mode", "loc", "map_val"]
            if config.get("loc_mode") == "geojson-id":
                map_props += ["geojson", "featureidkey"]
        for prop in map_props:
            final_data.append((prop, config.get(prop)))
    elif chart_type == "candlestick":
        group_by = config["cs_group"]
        for prop in ["cs_x", "cs_open", "cs_close", "cs_high", "cs_low"]:
            final_data.append((prop.split("_")[-1], config.get(prop)))
    elif chart_type == "treemap":
        group_by = config["treemap_group"]
        for prop in ["treemap_value", "treemap_label"]:
            final_data.append((prop.split("_")[-1], config.get(prop)))
    elif chart_type == "funnel":
        group_by = config["funnel_group"]
        for prop in ["funnel_value", "funnel_label"]:
            final_data.append((prop.split("_")[-1], config.get(prop)))
    elif chart_type == "clustergram":
        group_by = config["clustergram_group"]
        for prop in ["clustergram_value", "clustergram_label"]:
            final_data.append((prop.split("_")[-1], config.get(prop)))
    else:
        group_by = config.get("group")
        final_data.append(("X-Axis", config["x"]))
        y = make_list(config["y"])
        final_data.append(("Y-Axes" if len(y) > 1 else "Y-Axis", ",".join(y)))
        if chart_type in ZAXIS_CHARTS:
            final_data.append(("z", config.get("z")))
        if chart_type == "scatter" and config["trendline"]:
            final_data.append(("Trendline", "\u2714"))

    if group_by:
        group_by_cols = make_list(group_by)
        final_data.append(("Group By", ", ".join(group_by_cols)))
        group_type = config.get("group_type")
        final_data.append(("Group Type", group_type))
        if group_type == "bins":
            final_data.append(("Bin Type", config["bin_type"]))
            final_data.append(("Bins", config["bin_val"]))
        else:
            final_data.append(
                (
                    "Selected Groups",
                    ", ".join(
                        [
                            "({})".format(",".join([v[c] for c in group_by_cols]))
                            for v in make_list(config.get("group_val"))
                        ]
                    ),
                )
            )

    if config["cpg"]:
        final_data.append(("Chart Per Group", "\u2714"))
    if config["cpy"]:
        final_data.append(("Chart Per Y", "\u2714"))
    if chart_type in ANIMATION_CHARTS and config["animate"]:
        final_data.append(("Animate", "\u2714"))
    if chart_type in ANIMATE_BY_CHARTS and config["animate_by"]:
        final_data.append(("Animate By", ", ".join(make_list(config["animate_by"]))))

    return build_hoverable(
        html.I(className="ico-help-outline"),
        [
            html.B("Chart Configuration"),
            html.Ul(
                [
                    html.Li(
                        [html.B(text(prop)), html.Span(": {}".format(value))],
                        className="mb-0",
                    )
                    for prop, value in final_data
                    if value is not None
                ],
                className="mb-0",
            ),
        ],
        hover_class="saved-chart-config",
        top="unset",
    )


def init_callbacks(dash_app):
    def save_chart(*args):
        args = list(args)
        save_clicks = args.pop(0)
        current_deletes = [args.pop(0) or 0 for _ in range(MAX_SAVED_CHARTS)]
        inputs = args.pop(0)
        chart_inputs = args.pop(0)
        yaxis_data = args.pop(0)
        map_data = args.pop(0)
        cs_data = args.pop(0)
        treemap_data = args.pop(0)
        prev_save_clicks = args.pop(0)
        updated_configs = [args.pop(0) for _ in range(MAX_SAVED_CHARTS)]
        prev_deletes = [args.pop(0) or 0 for _ in range(MAX_SAVED_CHARTS)]

        delete_idx = None
        for i, (curr_delete, prev_delete) in enumerate(
            zip(current_deletes, prev_deletes)
        ):
            if curr_delete > prev_delete:
                delete_idx = i

        if delete_idx is None:
            if save_clicks == prev_save_clicks:
                raise DtalePreventUpdate

            config = dict_merge(
                inputs,
                chart_inputs,
                dict(yaxis=yaxis_data or {}),
                map_data,
                cs_data,
                treemap_data,
            )

            if not valid_chart(**config):
                raise DtalePreventUpdate

            for index, saved_config in enumerate(updated_configs):
                if saved_config is None:
                    updated_configs[index] = config
                    break
        else:
            updated_configs[delete_idx] = None

        return tuple([save_clicks] + updated_configs + current_deletes)

    def load_saved_chart(_ts, config, prev_config):
        if config == prev_config:
            raise DtalePreventUpdate
        if config is None:
            return dict(display="none"), None, None, None
        if is_app_root_defined(dash_app.server.config.get("APPLICATION_ROOT")):
            config["app_root"] = dash_app.server.config["APPLICATION_ROOT"]
        charts, _, _, _ = build_chart(**config)
        return dict(display="block"), charts, config, build_saved_header(config)

    dash_app.callback(
        [Output("save-clicks", "data")]
        + [Output("saved-chart-config-{}".format(i), "data") for i in SAVED_CHART_IDS]
        + [Output("saved-deletes-{}".format(i), "data") for i in SAVED_CHART_IDS],
        [Input("save-btn", "n_clicks")]
        + [Input("delete-saved-btn-{}".format(i), "n_clicks") for i in SAVED_CHART_IDS],
        [
            State("input-data", "data"),
            State("chart-input-data", "data"),
            State("yaxis-data", "data"),
            State("map-input-data", "data"),
            State("candlestick-input-data", "data"),
            State("treemap-input-data", "data"),
            State("save-clicks", "data"),
        ]
        + [State("saved-chart-config-{}".format(i), "data") for i in SAVED_CHART_IDS]
        + [State("saved-deletes-{}".format(i), "data") for i in SAVED_CHART_IDS],
    )(save_chart)

    for i in SAVED_CHART_IDS:
        dash_app.callback(
            [
                Output("saved-chart-div-{}".format(i), "style"),
                Output("saved-chart-{}".format(i), "children"),
                Output("prev-saved-chart-config-{}".format(i), "data"),
                Output("saved-chart-header-{}".format(i), "children"),
            ],
            [Input("saved-chart-config-{}".format(i), "modified_timestamp")],
            [
                State("saved-chart-config-{}".format(i), "data"),
                State("prev-saved-chart-config-{}".format(i), "data"),
            ],
        )(load_saved_chart)



import re

import pandas as pd

import dtale.global_state as global_state

from dtale.pandas_util import check_pandas_version, is_pandas3
from dtale.utils import format_data, get_bool_arg

# Patterns that could enable code execution via pandas.DataFrame.query()
_DANGEROUS_QUERY_PATTERNS = re.compile(
    r"(__\w+__"  # dunder attributes (__import__, __class__, etc.)
    r"|(?<!\w)import\s*\("  # import() calls
    r"|(?<!\w)exec\s*\("  # exec() calls
    r"|(?<!\w)eval\s*\("  # eval() calls
    r"|(?<!\w)compile\s*\("  # compile() calls
    r"|(?<!\w)open\s*\("  # open() calls
    r"|(?<!\w)getattr\s*\("  # getattr() calls
    r"|(?<!\w)setattr\s*\("  # setattr() calls
    r"|(?<!\w)delattr\s*\("  # delattr() calls
    r"|(?<!\w)globals\s*\("  # globals() calls
    r"|(?<!\w)locals\s*\("  # locals() calls
    r"|(?<!\w)vars\s*\("  # vars() calls
    r"|(?<!\w)dir\s*\("  # dir() calls
    r"|(?<!\w)os\."  # os module access
    r"|(?<!\w)sys\."  # sys module access
    r"|(?<!\w)subprocess"  # subprocess module access
    r"|(?<!\w)shutil\."  # shutil module access
    r")",
    re.IGNORECASE,
)


def validate_query_safety(query):
    """Defense-in-depth validation of query strings before passing to DataFrame.query().

    This catches dangerous patterns that could enable code execution, even if
    upstream filter construction is compromised.
    """
    if not query:
        return
    if _DANGEROUS_QUERY_PATTERNS.search(query):
        raise ValueError(
            "Query contains potentially unsafe content and has been blocked: {}".format(
                repr(query[:200])
            )
        )


def build_col_key(col):
    # Use backticks for pandas >= 0.25 to handle column names with spaces or protected words
    return "`{}`".format(col)


def build_query(data_id, query=None):
    curr_settings = global_state.get_settings(data_id) or {}
    return inner_build_query(curr_settings, query)


def inner_build_query(settings, query=None):
    query_segs = []
    for p in ["columnFilters", "outlierFilters"]:
        curr_filters = settings.get(p) or {}
        for col, filter_cfg in curr_filters.items():
            query_segs.append(filter_cfg["query"])
    if query not in [None, ""]:
        query_segs.append(query)
    joined_query_segs = " and ".join(query_segs)
    if joined_query_segs and settings.get("invertFilter", False):
        joined_query_segs = "~({})".format(joined_query_segs)
    return joined_query_segs


def load_index_filter(data_id):
    curr_settings = global_state.get_settings(data_id) or {}
    column_filters = curr_settings.get("columnFilters") or {}
    indexes = curr_settings.get("indexes", [])
    for col in indexes:
        cfg = column_filters.get(col)
        if cfg:
            start, end = (cfg.get(p) for p in ["start", "end"])
            if start and end:
                return {"date_range": [pd.Timestamp(start), pd.Timestamp(end)]}
            elif start:
                return {"date_range": [pd.Timestamp(start), None]}
            elif end:
                return {"date_range": [None, pd.Timestamp(end)]}
    return {}


def build_query_builder(data_id):
    from arcticdb import QueryBuilder

    curr_settings = global_state.get_settings(data_id) or {}
    q = QueryBuilder()
    result = inner_build_query_builder(curr_settings, q)
    if result is not None:
        return q[result]
    return None


def inner_build_query_builder(settings, query_builder):
    from dtale.column_filters import ArcticDBColumnFilter

    result = None
    for p in ["columnFilters", "outlierFilters"]:
        curr_filters = settings.get(p) or {}
        filter_items = list(curr_filters.items())
        idx = 0
        if len(filter_items):
            while result is None and idx < len(filter_items):
                col, filter_cfg = filter_items[idx]
                if col in settings.get("indexes", []):
                    idx += 1
                    continue
                fltr = ArcticDBColumnFilter(filter_cfg)
                result = fltr.builder.update_query_builder(query_builder)
                idx += 1
            for col, filter_cfg in filter_items[idx:]:
                if col in settings.get("indexes", []):
                    continue
                fltr = ArcticDBColumnFilter(filter_cfg)
                query_seg = fltr.builder.update_query_builder(query_builder)
                if query_seg is not None:
                    result &= query_seg

    if result is not None and settings.get("invertFilter", False):
        result = ~result
    return result


def run_query(
    df,
    query,
    context_vars=None,
    ignore_empty=False,
    pct=100,
    pct_type="random",
    stratified_group=None,
    highlight_filter=False,
):
    """
    Utility function for running :func:`pandas:pandas.DataFrame.query` . This function contains extra logic to
    handle when column names contain special characters.  Looks like pandas will be handling this in a future
    version: https://github.com/pandas-dev/pandas/issues/27017

    The logic to handle these special characters in the meantime is only available in Python 3+

    :param df: input dataframe
    :type df: :class:`pandas:pandas.DataFrame`
    :param query: query string
    :type query: str
    :param context_vars: dictionary of user-defined variables which can be referenced by name in query strings
    :type context_vars: dict, optional
    :param pct: random percentage of dataframe to load
    :type pct: int, optional
    :param stratified_group: the grouping to use when load a stratified random sample of dataframe
    :type stratified_group: str, optional
    :param highlight_filter: if true, then highlight which rows will be filtered rather than drop them
    :type highlight_filter: boolean, optional
    :return: filtered dataframe
    """

    def _load_pct(df):
        if pct is not None and pct < 100:
            if pct_type == "random":
                return df.sample(frac=pct / 100.0)
            elif pct_type == "stratified":
                sampling_rate = pct / 100.0
                num_rows = int((df.shape[0] * sampling_rate) // 1)
                num_classes = len(df[stratified_group].unique())
                num_rows_per_class = int(max(1, ((num_rows / num_classes) // 1)))
                if is_pandas3():
                    return (
                        df.groupby(stratified_group)
                        .apply(lambda x: x.sample(min(len(x), num_rows_per_class)))
                        .reset_index(-1, drop=True)
                        .reset_index()
                    )
                return df.groupby(stratified_group, group_keys=False).apply(
                    lambda x: x.sample(min(len(x), num_rows_per_class))
                )
            records = int(len(df) * (pct / 100.0))
            return df.head(records) if pct_type == "head" else df.tail(records)

        return df

    if (query or "") == "":
        if highlight_filter:
            return _load_pct(df), []
        return _load_pct(df)

    validate_query_safety(query)

    is_pandas25 = check_pandas_version("0.25.0")
    curr_app_settings = global_state.get_app_settings()
    engine = curr_app_settings.get("query_engine", "python")
    filtered_indexes = []
    if highlight_filter:
        filtered_indexes = set(
            df.query(
                query if is_pandas25 else query.replace("`", ""),
                local_dict=context_vars or {},
                engine=engine,
            ).index
        )
    else:
        df = df.query(
            query if is_pandas25 else query.replace("`", ""),
            local_dict=context_vars or {},
            engine=engine,
        )

    if not len(df) and not ignore_empty:
        raise Exception('query "{}" found no data, please alter'.format(query))

    if highlight_filter:
        return _load_pct(df), filtered_indexes
    return _load_pct(df)


def handle_predefined(data_id, df=None):
    import dtale.predefined_filters as predefined_filters

    df = global_state.get_data(data_id) if df is None else df
    filters = predefined_filters.get_filters()
    if not filters:
        return df

    curr_settings = global_state.get_settings(data_id) or {}
    filter_values = curr_settings.get("predefinedFilters")
    if not filter_values:
        return df

    for f in filters:
        if f.name in filter_values:
            filter_value = filter_values[f.name]
            if (
                filter_value.get("active", True)
                and filter_value.get("value") is not None
            ):
                df = f.handler(df, filter_value["value"])
    return df


def load_filterable_data(data_id, req, query=None, columns=None):
    filtered = get_bool_arg(req, "filtered")
    if global_state.is_arcticdb:
        query_builder = build_query_builder(data_id)
        instance = global_state.store.get(data_id)
        if filtered:
            data = instance.load_data(query_builder=query_builder, columns=columns)
        else:
            data = instance.load_data(columns=columns)
        data, _ = format_data(data)
        return data
    if filtered:
        final_query = query or build_query(data_id, global_state.get_query(data_id))
        return run_query(
            handle_predefined(data_id),
            final_query,
            global_state.get_context_variables(data_id),
            ignore_empty=True,
        )
    return global_state.get_data(data_id)



import inspect

from six import PY3

PREDEFINED_FILTERS = []


class PredefinedFilter(object):
    def __init__(
        self,
        name=None,
        description=None,
        handler=None,
        column=None,
        input_type=None,
        default=None,
        active=True,
    ):
        assert name is not None
        assert column is not None
        assert handler is not None and callable(handler)
        if PY3:
            assert len(inspect.signature(handler).parameters) == 2
        else:
            assert len(inspect.getargspec(handler).args) == 2
        assert input_type is not None
        assert input_type in ["input", "select", "multiselect"]

        self.name = name
        self.description = description
        self.handler = handler
        self.column = column
        self.input_type = input_type
        self.default = default
        self.active = active

    def asdict(self):
        return {
            "name": self.name,
            "description": self.description,
            "column": self.column,
            "inputType": self.input_type,
            "default": self.default,
            "active": self.active,
        }


def get_filters():
    return PREDEFINED_FILTERS


def set_filters(filters):
    global PREDEFINED_FILTERS

    PREDEFINED_FILTERS = [PredefinedFilter(**filter_spec) for filter_spec in filters]


def add_filters(filters):
    global PREDEFINED_FILTERS

    valid_filters = []
    for filter_spec in filters:
        new_filter = PredefinedFilter(**filter_spec)
        if next(
            (
                curr_filter
                for curr_filter in PREDEFINED_FILTERS
                if curr_filter.name == new_filter.name
            ),
            None,
        ):
            raise ValueError(
                "A predefined_filters filter already exists for {}".format(
                    new_filter.name
                )
            )
        valid_filters.append(new_filter)
    PREDEFINED_FILTERS += valid_filters


def init_filters():
    return {
        f.name: (
            dict(value=f.default, active=f.active)
            if f.default is not None
            else dict(active=f.active)
        )
        for f in PREDEFINED_FILTERS
    }



import pandas as pd

from dtale.app import show
from dtale.cli.clickutils import get_loader_options, handle_path, loader_prop_keys

"""
  IMPORTANT!!! These global variables are required for building any customized CLI loader.
  When build_loaders runs startup it will search for any modules containing the global variable LOADER_KEY.
"""
LOADER_KEY = "csv"
LOADER_PROPS = [
    dict(name="path", help="path to CSV file"),
    dict(name="proxy", help="proxy URL if you're passing in a URL for --csv-path"),
    dict(
        name="parse_dates",
        help="comma-separated string of column names which should be parsed as dates",
    ),
    dict(name="index_col", help="Column(s) to use as the row labels or the Datafame"),
    dict(name="delimiter", help="Delimiter to use (comma, tab, etc...)"),
]


# IMPORTANT!!! This function is required if you would like to be able to use this loader from the back-end.
def show_loader(**kwargs):
    return show(data_loader=lambda: loader_func(**kwargs), **kwargs)


def loader_func(**kwargs):
    path = handle_path(kwargs.pop("path"), kwargs)
    return pd.read_csv(
        path, **{k: v for k, v in kwargs.items() if k in loader_prop_keys(LOADER_PROPS)}
    )


# IMPORTANT!!! This function is required for building any customized CLI loader.
def find_loader(kwargs):
    """
    CSV implementation of data loader which will return a function if any of the
    `click` options based on LOADER_KEY & LOADER_PROPS have been used, otherwise return None

    :param kwargs: Optional keyword arguments to be passed from `click`
    :return: data loader function for CSV implementation
    """
    csv_opts = get_loader_options(LOADER_KEY, LOADER_PROPS, kwargs)
    if len([f for f in csv_opts.values() if f]):

        def _csv_loader():
            csv_arg_parsers = {  # TODO: add additional arg parsers
                "parse_dates": lambda v: v.split(",") if v else None
            }
            kwargs = {
                k: csv_arg_parsers.get(k, lambda v: v)(v) for k, v in csv_opts.items()
            }
            return loader_func(**kwargs)

        return _csv_loader
    return None



# AUTO GENERATED FILE - DO NOT EDIT

from dash.development.base_component import Component, _explicitize_args


class Wordcloud(Component):
    """A Wordcloud component.
    Wordcloud is a wrapper component for react-wordcloud usage by dash.
    It takes a property, `data`, containing a series of words and a series of weights and displays a wordcloud
    or group of wordclouds depending on whether a `group` value has been specified.

    Keyword arguments:

    - id (string; required):
        The ID used to identify this component in Dash callbacks.

    - data (dict; optional):
        Server-side data containing words \"data[group].x\" and weights
        \"data[group][y-prop]\".

    - group (list of strings; optional):
        List of properties to use as groups.

    - height (number; default 400):
        Height of wordcloud in pixels (default: 400).

    - y (list of strings; optional):
        List of properties to use as weights."""

    _children_props = []
    _base_nodes = ["children"]
    _namespace = "components"
    _type = "Wordcloud"

    @_explicitize_args
    def __init__(
        self,
        id=Component.REQUIRED,
        data=Component.UNDEFINED,
        y=Component.UNDEFINED,
        group=Component.UNDEFINED,
        height=Component.UNDEFINED,
        **kwargs
    ):
        self._prop_names = ["id", "data", "group", "height", "y"]
        self._valid_wildcard_attributes = []
        self.available_properties = ["id", "data", "group", "height", "y"]
        self.available_wildcard_properties = []
        _explicit_args = kwargs.pop("_explicit_args")
        _locals = locals()
        _locals.update(kwargs)  # For wildcard attrs and excess named props
        args = {k: _locals[k] for k in _explicit_args}

        for k in ["id"]:
            if k not in args:
                raise TypeError("Required argument `" + k + "` was not specified.")

        super(Wordcloud, self).__init__(**args)



import numpy as np
import pandas as pd
import requests
import string
import zipfile
from datetime import datetime

from six import BytesIO

import dtale.pandas_util as pandas_util


def covid():
    from dtale.cli.loaders.csv_loader import loader_func as load_csv

    data = load_csv(
        path="https://raw.githubusercontent.com/nytimes/covid-19-data/master/us-states.csv",
        parse_dates=["date"],
    )
    codes = load_csv(
        path="https://raw.githubusercontent.com/jasonong/List-of-US-States/master/states.csv"
    )
    codes = codes.set_index("State").to_dict()["Abbreviation"]
    data["state_code"] = data["state"].map(codes)
    return data, None


def seinfeld():
    from dtale.cli.loaders.csv_loader import loader_func as load_csv

    episodes = load_csv(
        path="https://github.com/4m4n5/the-seinfeld-chronicles/raw/master/episode_info.csv"
    )
    episodes = episodes[[c for c in episodes.columns if c not in ["Unnamed: 0"]]]
    scripts = load_csv(
        path="https://github.com/4m4n5/the-seinfeld-chronicles/raw/master/scripts.csv"
    )
    scripts = scripts[
        [c for c in scripts.columns if c not in ["Unnamed: 0", "Season", "EpisodeNo"]]
    ]
    return pd.merge(episodes, scripts, how="inner", on="SEID"), None


def load_zip(url):
    response = requests.get(url)
    with zipfile.ZipFile(BytesIO(response.content)) as thezip:
        for zipinfo in thezip.infolist():
            yield zipinfo.filename, thezip.open(zipinfo.filename)


def simpsons():
    from dtale.cli.loaders.csv_loader import loader_func as load_csv
    import dtale.global_state as global_state

    global_state.set_app_settings(dict(max_column_width=100, max_row_height=100))
    episodes = load_csv(
        path="https://github.com/aschonfeld/dtale-media/raw/master/datasets/simpsons_episodes.csv"
    )
    episodes = episodes.rename(columns={"id": "episode_id"})
    episodes.loc[:, "image_url"] = episodes["image_url"].apply(
        lambda x: "<img src='{}' style='height: auto; width: 100px;' />".format(x)
    )
    _, scripts = next(
        load_zip(
            "https://github.com/aschonfeld/dtale-media/raw/master/datasets/simpsons_script_lines.csv.zip"
        )
    )
    scripts = pd.read_csv(scripts)
    df = pd.merge(episodes, scripts, how="inner", on="episode_id")
    formatting = {"image_url": {"fmt": {"html": True}}}
    return df, {"columnFormats": formatting}


def video_games():
    _, games = next(
        load_zip(
            "https://github.com/aschonfeld/dtale-media/raw/master/datasets/vgsales.csv.zip"
        )
    )
    return pd.read_csv(games), None


def movies():
    _, movies = next(
        load_zip(
            "https://github.com/aschonfeld/dtale-media/raw/master/datasets/IMDb_movies.csv.zip"
        )
    )
    movies = pd.read_csv(movies)
    pandas_util.assign_col_data(
        movies,
        "year",
        (
            movies["year"]
            .where(~(movies["year"] == "TV Movie 2019"), "2019")
            .astype("int")
        ),
    )
    return movies, None


def time_dataframe():
    def series_data():
        if hasattr(np.random, "default_rng"):
            return np.random.default_rng(2).standard_normal(30)
        return np.random.randn(30)

    cols = string.ascii_uppercase[:4]
    data = {
        c: pd.Series(
            series_data(),
            index=pd.DatetimeIndex(
                pd.date_range(datetime(2000, 1, 1), periods=30, freq="B")
            ),
        )
        for c in cols
    }
    return pd.DataFrame(data), None



try:
    import streamlit.server.server as streamlit_server
except BaseException:  # ImportError, ModuleNotFoundError
    import streamlit.web.server.server as streamlit_server
import sys

try:
    from streamlit.cli import main
except BaseException:
    from streamlit.web.cli import main
from tornado.wsgi import WSGIContainer
from tornado.web import FallbackHandler

from dtale.app import build_app
from dtale.cli.clickutils import run

orig_start_listening = streamlit_server.start_listening


def _override_start_listening(app):
    dtale_app_obj = build_app(reaper_on=False)

    tr = WSGIContainer(dtale_app_obj)
    app.add_handlers(r".*", [(".*dtale.*", FallbackHandler, dict(fallback=tr))])
    orig_start_listening(app)


streamlit_server.start_listening = _override_start_listening

# we need to stop XSRF protection since dash won't work otherwise
additional_args = []
if "--server.enableCORS" not in sys.argv:
    additional_args += ["--server.enableCORS", "false"]
if "--server.enableXsrfProtection" not in sys.argv:
    additional_args += ["--server.enableXsrfProtection", "false"]
sys.argv += additional_args

if __name__ == "__main__":
    run(main)



import plotly.graph_objs as go
from dash.dependencies import Input, Output, State

from dtale.charts.utils import MAX_GROUPS
from dtale.dash_application.exceptions import DtalePreventUpdate


def init_callbacks(dash_app):
    def lock_zoom(clicks, relayout_data, figure):
        if not clicks:
            raise DtalePreventUpdate
        figure = go.Figure(figure)
        if relayout_data:
            figure.update_layout(scene_camera=relayout_data["scene.camera"])
        return figure

    for i in range(1, MAX_GROUPS + 1):
        dash_app.callback(
            Output("chart-{}".format(i), "figure"),
            [Input("lock-zoom-btn", "n_clicks")],
            [
                State("chart-{}".format(i), "relayoutData"),
                State("chart-{}".format(i), "figure"),
            ],
        )(lock_zoom)



from logging import getLogger

import click
import sys

from dtale.app import find_free_port, show
from dtale.cli.clickutils import LOG_LEVELS, get_log_options, run, setup_logging
from dtale.cli.loaders import check_loaders, setup_loader_options

logger = getLogger(__name__)


def validate_allow_cell_edits(ctx, param, value):
    if "--no-cell-edits" in sys.argv:
        return value
    return None


@click.command(name="main", help="Run D-Tale from command-line")
@click.option("--host", type=str, help="hostname or IP address of process")
@click.option("--port", type=int, help="port number of process")
@click.option("--debug", is_flag=True, help="flag to switch on Flask's debug mode")
@click.option(
    "--no-reaper",
    is_flag=True,
    help="flag to turn off auto-reaping (process cleanup after period of inactivity)",
)
@click.option(
    "--open-browser",
    is_flag=True,
    help="flag to automatically open default web browser on startup",
)
@click.option("--name", type=str, help="name to apply to your D-Tale session")
@click.option(
    "--no-cell-edits",
    is_flag=True,
    callback=validate_allow_cell_edits,
    help="flag to turn off cell editing in the grid",
)
@click.option(
    "--hide-shutdown", is_flag=True, help='flag to hide "Shutdown" button from users'
)
@click.option(
    "--github-fork",
    is_flag=True,
    help='flag to show "Fork Me On GitHub" link in upper right-hand corner of the app',
)
@click.option("--app-root", type=str)
@click.option(
    "--hide-drop-rows", is_flag=True, help='flag to hide "Drop Rows" button from users'
)
@click.option(
    "--theme",
    default="light",
    type=click.Choice(["light", "dark"]),
    help='theme you would like used for D-Tale, the default is "light"',
)
@click.option(
    "--precision",
    default=2,
    type=int,
    help="Default precision displayed by columns containing float data (default: 2)",
)
@click.option(
    "--show-columns",
    help="Comma-separated string of column names you would like displayed on load",
)
@click.option(
    "--hide-columns",
    help="Comma-separated string of column names you would like hidden on load",
)
@click.option(
    "--query-engine",
    default="python",
    type=click.Choice(["python", "numexpr"]),
    help='query engine you would like used for D-Tale, the default is "python"',
)
@click.option(
    "--sort",
    help=(
        "Comma-separated string of pipe-separated column names/sort directions you would like applied on load. "
        "EX: col1|ASC,col2|DESC"
    ),
)
@click.option(
    "--locked",
    help="Comma-separated string of column names you would like locked on the left-hand side of your grid on load",
)
@click.option(
    "--hide-header-editor",
    is_flag=True,
    help="flag to hide header editor when editing cells",
)
@setup_loader_options()
@click.option("--log", "logfile", help="Log file name")
@click.option(
    "--log-level",
    help="Set the logging level",
    type=click.Choice(list(LOG_LEVELS.keys())),
    default="info",
    show_default=True,
)
@click.option("-v", "--verbose", help="Set the logging level to debug", is_flag=True)
def main(
    host=None,
    port=None,
    debug=None,
    no_reaper=None,
    open_browser=False,
    name=None,
    no_cell_edits=None,
    hide_shutdown=None,
    github_fork=None,
    app_root=None,
    hide_drop_rows=None,
    **kwargs
):
    """
    Runs a local server for the D-Tale application.

    This local server is recommended when you have a pandas object stored in a CSV
    or retrievable from :class:`arctic.Arctic` data store.
    """
    log_opts = get_log_options(kwargs)
    setup_logging(
        log_opts.get("logfile"), log_opts.get("log_level"), log_opts.get("verbose")
    )

    data_loader = check_loaders(kwargs)

    # in order to handle the hierarchy of inputs if "--no-cell-edits" is not specified
    # then we'll update it to None
    allow_cell_edits = False if no_cell_edits is not None else None
    kwargs["show_columns"] = (
        kwargs["show_columns"].split(",") if kwargs.get("show_columns") else None
    )
    kwargs["hide_columns"] = (
        kwargs["hide_columns"].split(",") if kwargs.get("hide_columns") else None
    )
    kwargs["sort"] = (
        [tuple(sort.split("|")) for sort in kwargs["sort"].split(",")]
        if kwargs.get("sort")
        else None
    )
    kwargs["locked"] = kwargs["locked"].split(",") if kwargs.get("locked") else None
    show(
        host=host,
        port=int(port or find_free_port()),
        debug=debug,
        subprocess=False,
        data_loader=data_loader,
        reaper_on=not no_reaper,
        open_browser=open_browser,
        name=name,
        allow_cell_edits=allow_cell_edits,
        hide_shutdown=hide_shutdown,
        github_fork=github_fork,
        app_root=app_root,
        hide_drop_rows=hide_drop_rows,
        **kwargs
    )


if __name__ == "__main__":
    run(main)



import pandas as pd

from dtale.app import show
from dtale.cli.clickutils import get_loader_options, handle_path, loader_prop_keys

"""
  IMPORTANT!!! These global variables are required for building any customized CLI loader.
  When build_loaders runs startup it will search for any modules containing the global variable LOADER_KEY.
"""
LOADER_KEY = "excel"
LOADER_PROPS = [
    dict(name="path", help="path to Excel (.xls,xlsx) file"),
    dict(name="proxy", help="proxy URL if you're passing in a URL for --excel-path"),
    dict(
        name="parse_dates",
        help="comma-separated string of column names which should be parsed as dates",
    ),
    dict(name="index_col", help="Column(s) to use as the row labels or the Datafame"),
    dict(name="sheet", help="Name of the sheet to load from your Excel file"),
]


# IMPORTANT!!! This function is required if you would like to be able to use this loader from the back-end.
def show_loader(**kwargs):
    return show(data_loader=lambda: loader_func(**kwargs), **kwargs)


def load_file(sheet_name=None, **kwargs):
    path = kwargs.pop("path")
    engine = "xlrd" if path.endswith("xls") else "openpyxl"
    path = handle_path(path, kwargs)
    dfs = pd.read_excel(
        path,
        sheet_name=sheet_name,
        engine=engine,
        **{k: v for k, v in kwargs.items() if k in loader_prop_keys(LOADER_PROPS)}
    )
    if dfs is None or not len(dfs):
        raise Exception("Failed to load Excel file. Returned no data.")
    return dfs


def loader_func(**kwargs):
    sheet_name = kwargs.pop("sheet", None)
    dfs = load_file(sheet_name=sheet_name, **kwargs)
    if sheet_name:
        if sheet_name not in dfs:
            raise Exception(
                "Excel file loaded but there was no sheet named '{}'.".format(
                    sheet_name
                )
            )
        return dfs[sheet_name]
    # this is required because there is no support for loading multiple datasets at once from the CLI
    # I can add this later...
    return dfs[list(dfs.keys())[0]]


# IMPORTANT!!! This function is required for building any customized CLI loader.
def find_loader(kwargs):
    """
    Excel implementation of data loader which will return a function if any of the
    `click` options based on LOADER_KEY & LOADER_PROPS have been used, otherwise return None

    :param kwargs: Optional keyword arguments to be passed from `click`
    :return: data loader function for CSV implementation
    """
    excel_opts = get_loader_options(LOADER_KEY, LOADER_PROPS, kwargs)
    if len([f for f in excel_opts.values() if f]):

        def _excel_loader():
            excel_arg_parsers = {  # TODO: add additional arg parsers
                "parse_dates": lambda v: v.split(",") if v else None
            }
            kwargs = {
                k: excel_arg_parsers.get(k, lambda v: v)(v)
                for k, v in excel_opts.items()
            }
            return loader_func(**kwargs)

        return _excel_loader
    return None



from six import PY3

from dtale.app import show
from dtale.cli.clickutils import get_loader_options

"""
  IMPORTANT!!! These global variables are required for building any customized CLI loader.
  When build_loaders runs startup it will search for any modules containing the global variable LOADER_KEY.
"""
LOADER_KEY = "r"
LOADER_PROPS = [
    dict(name="path", help="path to R dataset file"),
    dict(name="dataset", help="the name of the dataset within your file"),
]


# IMPORTANT!!! This function is required if you would like to be able to use this loader from the back-end.
def show_loader(**kwargs):
    return show(data_loader=lambda: loader_func(**kwargs), **kwargs)


def loader_func(**kwargs):
    if not PY3:
        raise EnvironmentError("In order to use the R you must be on python3!")
    try:
        import rpy2.robjects as ro
        from rpy2.robjects.packages import importr
        from rpy2.robjects import pandas2ri

        from rpy2.robjects.conversion import localconverter
    except ImportError:
        raise ImportError("In order to use the R loader you must install rpy2!")

    base = importr("base")
    path = kwargs.pop("path")
    base.load(path)
    with localconverter(ro.default_converter + pandas2ri.converter):
        dataset = kwargs.pop("dataset", None) or base.ls()[0]
        df = ro.conversion.rpy2py(base.get(dataset))
        return df


# IMPORTANT!!! This function is required for building any customized CLI loader.
def find_loader(kwargs):
    """
    R dataset implementation of data loader which will return a function if any of the
    `click` options based on LOADER_KEY & LOADER_PROPS have been used, otherwise return None

    :param kwargs: Optional keyword arguments to be passed from `click`
    :return: data loader function for R dataset implementation
    """
    r_opts = get_loader_options(LOADER_KEY, LOADER_PROPS, kwargs)
    if len([f for f in r_opts.values() if f]):

        def _r_loader():
            return loader_func(**r_opts)

        return _r_loader
    return None
